import {Point, Rect, Size} from "./engine/datatypes.ts";
import {ALL_STATS, Stat} from "./datatypes.ts";
import {DrawPile} from "./drawpile.ts";
import {D} from "./engine/public.ts";
import {sprDrips, sprRaccoonWalking, sprResourcePickup, sprStatPickup} from "./sprites.ts";
import {BG_INSET, FG_TEXT} from "./colors.ts";
import {getPlayerProgress} from "./playerprogress.ts";
import {Architecture, CellView, LoadedNewMap} from "./newmap.ts";


export class HuntMode {
  map: LoadedNewMap
  player: Point

  drawpile: DrawPile
  frame: number
  depth: number

  constructor(depth: number, map: LoadedNewMap) {
    this.map = map;
    this.player = map.entrance;

    this.drawpile = new DrawPile();
    this.frame = 0;
    this.depth = depth;
    this.#updateVisibilityAndPossibleMoves();
  }

  getDepth() {
    return this.depth;
  }

  // == update logic ==
  #updateVisibilityAndPossibleMoves() {
    let revealAt = (depth: number, xStart: number, yStart: number) => {
      let cell = this.map.get(new Point(xStart, yStart));
      cell.revealed = true;
      if (depth <= 0 || cell.architecture == Architecture.Wall) {
        return;
      }
      for (let dx = -1; dx <= 1; dx++) {
        for (let dy = -1; dy <= 1; dy++) {
          let position = new Point(xStart + dx, yStart + dy);
          revealAt(depth - 1, position.x, position.y);
        }

      }
    }
    // NOTE: Depth 1 to reveal slightly less
    revealAt(2, this.player.x, this.player.y);
  }

  #collectResources() {
    let cell = this.map.get(this.player);

    /*
    if (present.content.type == "stairs") {
      getPlayerProgress().addBlood(1000);
      initHuntMode(new HuntMode(this.depth + 1));
    }
     */

    let pickup = cell.pickup;
    if (pickup != null) {
      switch (pickup) {
        case "AGI":
        case "INT":
        case "CHA":
        case "PSI":
          getPlayerProgress().add(pickup, 1);
          getPlayerProgress().purloinItem();
          break;
        case "EXP":
          getPlayerProgress().addExperience(25);
          getPlayerProgress().purloinItem();
          break;
        default:
          throw `not sure how to handle ${pickup}`
      }
      cell.pickup = null;
    }
  }

  #computeCostToMoveTo(mapPosition: Point): number | null {
    let present = this.map.get(mapPosition);

    if (present.architecture != Architecture.Floor) {
      return null;
    }

    let dist = Math.max(
      Math.abs(mapPosition.x - this.player.x),
      Math.abs(mapPosition.y - this.player.y)
    );

    if (dist != 1) { return null; }

    let pickup = present.pickup;
    if (pickup == null) { return 10; }
    return 100;  // any other pickup (EXP, stats, etc)
  }

  movePlayerTo(newPosition: Point) {
    this.player = newPosition;
    this.#updateVisibilityAndPossibleMoves();
    this.#collectResources();
  }

  // draw
  update() {
    this.frame += 1;
    this.drawpile.clear();

    let globalOffset =
      new Point(this.player.x * MAP_CELL_ONSCREEN_SIZE.w, this.player.y * MAP_CELL_ONSCREEN_SIZE.h).offset(
        new Point(-192, -192)
      )

    for (let y = 0; y < this.map.size.h; y += 1) {
      for (let x = 0; x < this.map.size.w; x += 1) {
        let cellOffset = new Point(x * MAP_CELL_ONSCREEN_SIZE.w, y * MAP_CELL_ONSCREEN_SIZE.h).offset(globalOffset.negate());
        let cell = this.map.get(new Point(x, y))
        let belowIsBlock = true;
        if (y < this.map.size.h - 1) {
          let below = this.map.get(new Point(x, y + 1));
          belowIsBlock = !below.revealed || below.architecture == Architecture.Wall;
        }

        this.#drawMapCell(cellOffset, new Point(x, y), cell, belowIsBlock);
      }
    }
    this.#drawPlayer(globalOffset);

    this.drawpile.executeOnClick();
  }

  draw() {
    this.drawpile.draw()
  }

  #drawMapCell(
    cellOffset: Point,
    mapPosition: Point,
    cellData: CellView,
    belowIsBlock: boolean
  ) {
    const OFFSET_FLOOR = -256;
    const OFFSET_AIR = 0;
    const depth = mapPosition.y;
    const onFloor = OFFSET_FLOOR + depth;
    const inAir = OFFSET_AIR + depth;

    if (!cellData.revealed) {
      return;
    }

    let cellTopLeft = cellOffset.offset(new Size(-MAP_CELL_ONSCREEN_SIZE.w / 2, -MAP_CELL_ONSCREEN_SIZE.h / 2));
    let cellSize = MAP_CELL_ONSCREEN_SIZE;

    if (cellData.architecture == Architecture.Wall) {
      if (!belowIsBlock) {
        this.drawpile.add(inAir, () => {
          D.drawSprite(sprDrips, cellOffset.offset(new Point(0, -cellSize.h / 2)), 1, {xScale: 3, yScale: 3})
        })
      }
      return;
    }

    // draw inset zone
    let cost = this.#computeCostToMoveTo(mapPosition);
    this.drawpile.addClickable(onFloor,
      (hover: boolean) => {
        D.fillRect(cellTopLeft, cellSize, hover ? FG_TEXT : BG_INSET)

        /*
        // TODO: Stairs
        if (cellData.content.type == "stairs") {
          // draw ladder if applicable
          D.drawSprite(sprLadder, cellTopLeft, 0, {xScale: 3, yScale: 3});
        }
         */
      },
      new Rect(cellTopLeft, cellSize),
      cost != null && cost <= getPlayerProgress().getBlood(),
      () => {
        if (cost != null) {
          getPlayerProgress().spendBlood(cost);
          this.movePlayerTo(mapPosition)
        }
      }
    );


    if (belowIsBlock) {
      // draw the underhang
      this.drawpile.add(onFloor, () => {
        D.drawSprite(sprDrips, cellOffset.offset(new Point(0, cellSize.h / 2)), 0, {xScale: 3, yScale: 3})
      })
    }

    let pickup = cellData.pickup;
    if (pickup != null) {
      let statIndex = ALL_STATS.indexOf(pickup as Stat);
      if (statIndex != -1) {
        let extraXOffset = 0; // Math.cos(this.frame / 80 + mapPosition.x + mapPosition.y) * 1;
        let extraYOffset = Math.sin(this.frame / 50 + mapPosition.x * 2 + mapPosition.y * 0.75) * 6 - 18;
        this.drawpile.add(inAir, () => {
          D.drawSprite(
            sprStatPickup,
            cellOffset.offset(new Point(extraXOffset, extraYOffset)),
            statIndex,
            {
              xScale: 3,
              yScale: 3,
            }
          )
        });
      }

      if (pickup == "EXP") {
        this.drawpile.add(inAir, () => {
          D.drawSprite(
            sprResourcePickup,
            cellOffset.offset(new Point(0, -16 * 3)),
            0,
            {
              xScale: 3,
              yScale: 3,
            }
          );
        });
      }
    }
  }

  #drawPlayer(globalOffset: Point) {
    let cellOffset = new Point(
      this.player.x * MAP_CELL_ONSCREEN_SIZE.w,
      this.player.y * MAP_CELL_ONSCREEN_SIZE.h
    ).offset(globalOffset.negate())
    this.drawpile.add(this.player.y, () => {
      D.drawSprite(
        sprRaccoonWalking,
        cellOffset.offset(new Point(0, 22)),
        0, {
          xScale: 3,
          yScale: 3
        }
      )
    });
  }
}

const MAP_CELL_ONSCREEN_SIZE: Size = new Size(96, 48)

let active: HuntMode | null = null;
export function initHuntMode(huntMode: HuntMode) {
  active = huntMode;
}

export function getHuntMode() {
  if (active == null) {
    throw `trying to get player progress before it has been initialized`
  }
  return active;
}
